"""
Django settings for simqsvr project.

Generated by 'django-admin startproject' using Django 3.2.25.

For more information on this file, see
https://docs.djangoproject.com/en/3.2/topics/settings/

For the full list of settings and their values, see
https://docs.djangoproject.com/en/3.2/ref/settings/
"""

import os
import logging
from pathlib import Path
from simqsvr.version import VERSION

_logger = logging.getLogger(__name__)

# Build paths inside the project like this: BASE_DIR / 'subdir'.
BASE_DIR = Path(__file__).resolve().parent.parent


# Quick-start development settings - unsuitable for production
# See https://docs.djangoproject.com/en/3.2/howto/deployment/checklist/

# SECURITY WARNING: keep the secret key used in production secret!
SECRET_KEY = "Y536wJqNpbSx3szOj5UyNQKiKdHNaIBbZBxfBeHAenzsD61cUvT5ZIP9JUX7A1BBTMrEky9bbeWGSSZYfJWGoybB101bVaVrhu5zGnYvS4tfat2pMwLy3a8O"

# SECURITY WARNING: don't run with debug turned on in production!
DEBUG = True
ALLOWED_HOSTS = [
    "localhost",
    "127.0.0.1",
]


# Application definition

INSTALLED_APPS = [
    "django_apis",
    "simqsvr",
    "django.contrib.admin",
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "django.contrib.messages",
    "django.contrib.staticfiles",
]

MIDDLEWARE = [
    "django.middleware.security.SecurityMiddleware",
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.middleware.common.CommonMiddleware",
    "django.middleware.csrf.CsrfViewMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
    "django.middleware.clickjacking.XFrameOptionsMiddleware",
]

ROOT_URLCONF = "simqsvr.urls"

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "DIRS": [],
        "APP_DIRS": True,
        "OPTIONS": {
            "context_processors": [
                "django.template.context_processors.debug",
                "django.template.context_processors.request",
                "django.contrib.auth.context_processors.auth",
                "django.contrib.messages.context_processors.messages",
            ],
        },
    },
]

WSGI_APPLICATION = "simqsvr.wsgi.application"


# Database
# https://docs.djangoproject.com/en/3.2/ref/settings/#databases
DEFAULT_DATABASE = os.path.abspath("database/db.sqlite3")
try:
    os.makedirs(os.path.dirname(DEFAULT_DATABASE), exist_ok=True)
except Exception as error:
    _logger.warning(
        "Create DEFAULT_DATABASE root failed: root=%s",
        os.path.dirname(DEFAULT_DATABASE),
    )

DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.sqlite3",
        "NAME": DEFAULT_DATABASE,
    }
}


# Password validation
# https://docs.djangoproject.com/en/3.2/ref/settings/#auth-password-validators

AUTH_PASSWORD_VALIDATORS = [
    {
        "NAME": "django.contrib.auth.password_validation.UserAttributeSimilarityValidator",
    },
    {
        "NAME": "django.contrib.auth.password_validation.MinimumLengthValidator",
    },
    {
        "NAME": "django.contrib.auth.password_validation.CommonPasswordValidator",
    },
    {
        "NAME": "django.contrib.auth.password_validation.NumericPasswordValidator",
    },
]


# Internationalization
# https://docs.djangoproject.com/en/3.2/topics/i18n/

LANGUAGE_CODE = "zh-hans"

TIME_ZONE = "Asia/Shanghai"

USE_I18N = True

USE_L10N = True

USE_TZ = True


# Static files (CSS, JavaScript, Images)
# https://docs.djangoproject.com/en/3.2/howto/static-files/

STATIC_URL = "simq/admin/static/"
STATIC_ROOT = "/www/static/simq/admin/static/"
MEDIA_URL = "simq/admin/upload/"
MEDIA_ROOT = "/www/upload/simq/admin/upload/"


# Default primary key field type
# https://docs.djangoproject.com/en/3.2/ref/settings/#default-auto-field

DEFAULT_AUTO_FIELD = "django.db.models.BigAutoField"
DATA_UPLOAD_MAX_NUMBER_FIELDS = 200000
SESSION_COOKIE_NAME = "simqsvr_session_cookie_id"
CSRF_COOKIE_NAME = "simqsvr_csrf_cookie_id"
CSRF_COOKIE_HTTPONLY = True
LOGGING = {
    "version": 1,
    "disable_existing_loggers": False,
    "handlers": {
        "console": {
            "class": "logging.StreamHandler",
        },
    },
    "root": {
        "handlers": ["console"],
        "level": "INFO",
    },
    "loggers": {
        "django": {
            "handlers": ["console"],
            "level": "INFO",
            "propagate": False,
        },
        "django_apis": {
            "handlers": ["console"],
            "level": "INFO",
            "propagate": False,
        },
        "simq": {
            "handlers": ["console"],
            "level": "INFO",
            "propagate": False,
        },
        "simqsvr": {
            "handlers": ["console"],
            "level": "INFO",
            "propagate": False,
        },
    },
}


DJANGO_APIS_OPENAPI_TITLE = "simqsvr"
DJANGO_APIS_OPENAPI_DESCRIPTION = "simqsvr"
DJANGO_APIS_OPENAPI_VERSION = VERSION
DJANGO_APIS_OPENAPI_SECURITY_DEFINITIONS = {
    "username": {
        "type": "apiKey",
        "name": "TokenUsername",
        "in": "header",
    },
    "name": {
        "type": "apiKey",
        "name": "TokenUserRealName",
        "in": "header",
    },
    "department": {
        "type": "apiKey",
        "name": "TokenDepartmentName",
        "in": "header",
    },
}


from django_environ_settings import django_environ_settings_patch_all

django_environ_settings_patch_all()

try:
    from localsettings import *
except ImportError:
    pass

from django_app_requires import patch_all as django_app_requires_patch_all

django_app_requires_patch_all()

try:
    os.makedirs(STATIC_ROOT, exist_ok=True)
except Exception as error:
    _logger.warning(
        "Create STATIC_ROOT failed: STATIC_ROOT=%s, error=%s",
        STATIC_ROOT,
        error,
    )

try:
    os.makedirs(MEDIA_ROOT, exist_ok=True)
except Exception as error:
    _logger.warning(
        "Create MEDIA_ROOT failed: MEDIA_ROOT=%s, error=%s",
        MEDIA_ROOT,
        error,
    )



import uuid
import time
import json
import datetime

from django_redis import get_redis_connection
from django_apis.exceptions import Forbidden

from .settings import SIMQ_ACK_EVENT_EXPIRE
from .settings import SIMQ_DONE_ITEM_EXPIRE
from .settings import SIMQ_MQ_DEFAULT
from .settings import SIMQ_MQS
from .settings import SIMQ_PREFIX
from .settings import SIMQ_APIKEYS
from .settings import SIMQ_APIKEY_HEADER


def apikey_validate(request):
    if SIMQ_APIKEYS:
        if request.META.get(SIMQ_APIKEY_HEADER, None) in SIMQ_APIKEYS:
            return
        raise Forbidden()


def getdb(channel):
    name = SIMQ_MQS.get(channel, SIMQ_MQ_DEFAULT)
    return get_redis_connection(name)


def hset(db, key, data):
    mapping = {}
    for item, value in data.items():
        mapping[item] = json.dumps(value)
    db.hset(key, mapping=mapping)


def hgetall(db, key):
    data = {}
    mapping = db.hgetall(key)
    for item, value in mapping.items():
        data[item] = json.loads(value)
    return data


def get_mq_key(channel):
    return f"{SIMQ_PREFIX}:mq:{channel}"


def get_tmp_key(channel, tmpid):
    return f"{SIMQ_PREFIX}:tmp:{channel}:{tmpid}"


def get_done_key(id):
    if isinstance(id, bytes):
        id = id.decode("utf-8")
    return f"{SIMQ_PREFIX}:done:{id}"


def get_item_key(id):
    if isinstance(id, bytes):
        id = id.decode("utf-8")
    return f"{SIMQ_PREFIX}:item:{id}"


def _dumps(v):
    if isinstance(v, datetime.datetime):
        return v.isoformat()
    return v


def get_msgdata(msg):
    return json.dumps(msg, ensure_ascii=False, default=_dumps)


def rpush(channel, data, id=None):
    id = id or str(uuid.uuid4())
    db = getdb(channel)
    qkey = get_mq_key(channel)
    ikey = get_item_key(id=id)
    msg = {
        "id": id,
        "channel": channel,
        "add_time": time.time(),
        "mod_time": time.time(),
        "status": "ready",
        "data": data,
    }
    hset(db, ikey, msg)
    db.rpush(qkey, id)
    return id


def lpush(channel, data, id=None):
    id = id or str(uuid.uuid4())
    db = getdb(channel)
    qkey = get_mq_key(channel)
    ikey = get_item_key(id=id)
    msg = {
        "id": id,
        "channel": channel,
        "add_time": time.time(),
        "mod_time": time.time(),
        "status": "ready",
        "data": data,
    }
    hset(db, ikey, msg)
    db.lpush(qkey, id)
    return id


def pop(channel, worker, timeout=5):
    tmpid = str(uuid.uuid4())
    db = getdb(channel=channel)
    qkey = get_mq_key(channel=channel)
    tkey = get_tmp_key(channel=channel, tmpid=tmpid)
    id = db.brpoplpush(qkey, tkey, timeout=timeout)
    if id is None:
        return None
    ikey = get_item_key(id=id)
    msg = hgetall(db, ikey)
    if not msg:
        return None
    # 处理被取消的消息
    if msg.get("cancel_flag", False):
        hset(
            db,
            ikey,
            {
                "cancel_status": "canceled",
                "canceled_time": time.time(),
                "status": "canceled",
                "mod_time": time.time(),
                "done_time": time.time(),
                "result": None,
            },
        )
        tkey = get_tmp_key(channel=channel, tmpid=tmpid)
        db.delete(tkey)
        dkey = get_done_key(id=msg["id"])
        db.lpush(dkey, msg["id"])
        db.expire(dkey, SIMQ_ACK_EVENT_EXPIRE)
        db.expire(ikey, SIMQ_DONE_ITEM_EXPIRE)
        return None
    # 更新消息状态
    hset(
        db,
        ikey,
        {
            "running_id": tmpid,
            "start_time": time.time(),
            "mod_time": time.time(),
            "status": "running",
            "worker": worker,
        },
    )
    # 返回消息体
    msg = hgetall(db, ikey)
    return msg


def ack(channel, id, result=None):
    db = getdb(channel=channel)
    ikey = get_item_key(id=id)
    msg = hgetall(db, ikey)
    if not msg:
        return None
    hset(
        db,
        ikey,
        {
            "ack_time": time.time(),
            "mod_time": time.time(),
            "status": "done",
            "result": result,
        },
    )
    tkey = get_tmp_key(channel=channel, tmpid=msg["running_id"])
    db.delete(tkey)
    dkey = get_done_key(id=msg["id"])
    db.lpush(dkey, msg["id"])
    db.expire(dkey, SIMQ_ACK_EVENT_EXPIRE)
    db.expire(ikey, SIMQ_DONE_ITEM_EXPIRE)
    return True


def query(channel, id, timeout=0):
    db = getdb(channel=channel)
    ikey = get_item_key(id=id)
    if timeout == 0:
        msg = hgetall(db, ikey)
        if not msg:
            return None
        return msg
    else:
        dkey = get_done_key(id=id)
        db.brpop(dkey, timeout=timeout)
        msg = hgetall(db, ikey)
        if not msg:
            return None
        return msg


def cancel(channel, id):
    db = getdb(channel=channel)
    ikey = get_item_key(id=id)
    msg = hgetall(db, ikey)
    if not msg:
        return False
    if msg["status"] in ["running", "done", "canceled"]:
        return False
    hset(
        db,
        ikey,
        {
            "cancel_status": "canceling",
            "cancel_flag": True,
            "cancel_time": time.time(),
        },
    )
    return True


def ret(channel, id):
    db = getdb(channel=channel)
    ckey = get_mq_key(channel=channel)
    ikey = get_item_key(id=id)
    msg = hgetall(db, ikey)
    # 没有消息体，无法return
    if not msg:
        return False
    # 状态不为running，无法return
    if msg.get("status", "unknown") != "running":
        return False
    # 如果存在running状态变量，则删除
    if "running_id" in msg:
        tkey = get_tmp_key(channel=channel, tmpid=msg["running_id"])
        db.delete(tkey)
    # 更新消息状态
    hset(
        db,
        ikey,
        {
            "status": "ready",
            "return_flag": True,
            "return_time": time.time(),
            "mod_time": time.time(),
            "return_count": msg.get("return_count", 0) + 1,
            "worker": None,
        },
    )
    db.lpush(ckey, id)
    return True



from django.urls import path
from . import views

urlpatterns = [
    path("<str:channel>/rpush", views.rpush),
    path("<str:channel>/lpush", views.lpush),
    path("<str:channel>/pop", views.pop),
    path("<str:channel>/ack", views.ack),
    path("<str:channel>/ret", views.ret),
    path("<str:channel>/query", views.query),
    path("<str:channel>/cancel", views.cancel),
    # path("<str:channel>/delete", views.delete),
    # path("backup", views.backup),
]



from typing import Optional
from typing import Any

import pydantic
from django_apis.views import apiview

from . import services


class PushPayload(pydantic.BaseModel):
    id: Optional[str] = None
    data: Any


@apiview(methods="post", tags="simq")
def rpush(request, channel: str, payload: PushPayload) -> str:
    """添加消息（高优先级）"""
    services.apikey_validate(request=request)
    return services.rpush(
        channel=channel,
        data=payload.data,
        id=payload.id,
    )


@apiview(methods="post", tags="simq")
def lpush(request, channel: str, payload: PushPayload) -> str:
    """添加消息（低优先级）"""
    services.apikey_validate(request=request)
    return services.lpush(
        channel=channel,
        data=payload.data,
        id=payload.id,
    )


class PopPayload(pydantic.BaseModel):
    worker: str
    timeout: int = 5


@apiview(methods="post", tags="simq")
def pop(request, channel: str, payload: PopPayload) -> Any:
    """获取消息执行"""
    services.apikey_validate(request=request)
    return services.pop(
        channel=channel,
        timeout=payload.timeout,
        worker=payload.worker,
    )


class AckPayload(pydantic.BaseModel):
    id: str
    result: Optional[Any] = None


@apiview(methods="post", tags="simq")
def ack(request, channel: str, payload: AckPayload) -> bool:
    """确认消息执行完成"""
    services.apikey_validate(request=request)
    return services.ack(
        channel=channel,
        id=payload.id,
        result=payload.result,
    )


class QueryPayload(pydantic.BaseModel):
    id: str
    timeout: int = 0


@apiview(methods="post", tags="simq")
def query(request, channel: str, payload: QueryPayload) -> bool:
    """查询消息详情。

    timeout: 为0时，表示及时返回当前任务详情。不为0时，等待消息确认或超时后返回最新任务详情。
    """
    services.apikey_validate(request=request)
    return services.query(
        channel=channel,
        id=payload.id,
        timeout=payload.timeout,
    )


class CancelPayload(pydantic.BaseModel):
    id: str


@apiview(methods="post", tags="simq")
def cancel(request, channel: str, payload: CancelPayload) -> bool:
    """取消消息（仅限待处理状态）。"""
    services.apikey_validate(request=request)
    return services.cancel(
        channel=channel,
        id=payload.id,
    )


class RetPayload(pydantic.BaseModel):
    id: str


@apiview(methods="post", tags="simq")
def ret(request, channel: str, payload: RetPayload) -> bool:
    """退还消息"""
    services.apikey_validate(request=request)
    return services.ret(
        channel=channel,
        id=payload.id,
    )



from django.conf import settings


SIMQ_MQ_DEFAULT = getattr(settings, "SIMQ_MQ_DEFAULT", "default")
SIMQ_MQS = getattr(settings, "SIMQ_MQS", {})
SIMQ_PREFIX = getattr(settings, "SIMQ_PREFIX", "simq")

# 已完成消息通知事件默认保留10分钟
SIMQ_ACK_EVENT_EXPIRE = getattr(settings, "SIMQ_ACK_EVENT_EXPIRE", 10 * 60)

# 已完成消息及结果默认保留7天
SIMQ_DONE_ITEM_EXPIRE = getattr(settings, "SIMQ_DONE_ITEM_EXPIRE", 60 * 60 * 7)

# 接口认证
SIMQ_APIKEYS = getattr(settings, "SIMQ_APIKEYS", [])
SIMQ_APIKEY_HEADER = getattr(settings, "SIMQ_APIKEY_HEADER", "HTTP_APIKEY")



# -*- coding: utf-8 -*-
import os
from io import open
from setuptools import setup
from setuptools import find_packages

here = os.path.abspath(os.path.dirname(__file__))

with open(os.path.join(here, "README.md"), "r", encoding="utf-8") as fobj:
    long_description = fobj.read()

with open(os.path.join(here, "requirements.txt"), "r", encoding="utf-8") as fobj:
    requires = [x.strip() for x in fobj.readlines() if x.strip()]


setup(
    name="simqsvr",
    version="0.1.1",
    description="SIMQ服务器。基于redis的消息队列，通过web接口提供对外服务。",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Chen Han",
    maintainer="Chen Han",
    license="Apache License, Version 2.0",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3 :: Only",
    ],
    keywords=["django"],
    install_requires=requires,
    packages=find_packages(".", exclude=["simqtest", "simqtest.migrations"]),
    py_modules=["manage_simqsvr"],
    zip_safe=False,
    include_package_data=True,
    entry_points={
        "console_scripts": [
            "manage-simqsvr = manage_simqsvr:main",
        ]
    },
)
